@@ -4,38 +4,40 @@ import datetime
 import logging
 import time
 import uuid
+from collections import defaultdict
 from email.header import Header
 from email.mime.multipart import MIMEMultipart
 from email.mime.text import MIMEText
 from email.utils import formataddr, getaddresses
 from urllib.parse import (
-    urlsplit, parse_qsl, urlencode, urlunsplit, quote, urljoin)
+    parse_qsl, quote, urlencode, urljoin, urlsplit, urlunsplit)
 
 try:
     import html2text
 except ImportError:
     html2text = None
 
+from genshi.core import END, START, Attrs, QName
+from genshi.template import MarkupTemplate, TextTemplate
 from sql import Literal
 from sql.aggregate import Count
-from sql.functions import Substring, Position
-from genshi.template import MarkupTemplate, TextTemplate
-from genshi.core import START, END, QName, Attrs
 
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.model import (fields, ModelSQL, ModelView, Workflow, Unique,
-    EvalEnvironment, dualmethod)
+from trytond.model import (
+    EvalEnvironment, Index, ModelSQL, ModelView, Unique, Workflow, dualmethod,
+    fields)
 from trytond.pool import Pool
-from trytond.pyson import PYSONDecoder, Eval, If
+from trytond.pyson import Eval, If, PYSONDecoder, TimeDelta
 from trytond.report import Report
-from trytond.sendmail import sendmail_transactional, SMTPDataManager
+from trytond.sendmail import SMTPDataManager, sendmail_transactional
 from trytond.tools import grouped_slice, reduce_ids
+from trytond.tools.email_ import convert_ascii_email, set_from_header
 from trytond.transaction import Transaction
 from trytond.url import http_host
 from trytond.wsgi import Base64Converter
 
-from .exceptions import DomainError, ConditionError, TemplateError
+from .exceptions import ConditionError, DomainError, TemplateError
 from .mixin import MarketingAutomationMixin
 
 if not config.get(
@@ -53,14 +55,14 @@ logger = logging.getLogger(__name__)
 def _formataddr(name, email):
     if name:
         name = str(Header(name, 'utf-8'))
-    return formataddr((name, email))
+    return formataddr((name, convert_ascii_email(email)))
 
 
 class Scenario(Workflow, ModelSQL, ModelView):
     "Marketing Scenario"
     __name__ = 'marketing.automation.scenario'
 
-    name = fields.Char("Name")
+    name = fields.Char("Name", translate=True)
     model = fields.Selection('get_models', "Model", required=True)
     domain = fields.Char(
         "Domain", required=True,
@@ -71,15 +73,23 @@ class Scenario(Workflow, ModelSQL, ModelView):
         fields.Integer("Records"), 'get_record_count')
     record_count_blocked = fields.Function(
         fields.Integer("Records Blocked"), 'get_record_count')
+    unsubscribable = fields.Boolean(
+        "Unsubscribable",
+        help="If checked parties are also unsubscribed from the scenario.")
     state = fields.Selection([
             ('draft', "Draft"),
             ('running', "Running"),
             ('stopped', "Stopped"),
-            ], "State", required=True, readonly=True)
+            ], "State", required=True, readonly=True, sort=False)
 
     @classmethod
     def __setup__(cls):
         super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t, (t.state, Index.Equality()),
+                where=t.state.in_(['draft', 'running'])))
         cls._transitions |= set((
                 ('draft', 'running'),
                 ('running', 'stopped'),
@@ -107,6 +117,10 @@ class Scenario(Workflow, ModelSQL, ModelView):
     def default_domain(cls):
         return '[]'
 
+    @classmethod
+    def default_unsubscribable(cls):
+        return False
+
     @classmethod
     def get_models(cls):
         pool = Pool()
@@ -155,13 +169,15 @@ class Scenario(Workflow, ModelSQL, ModelView):
         return count
 
     @classmethod
-    def validate(cls, scenarios):
-        super().validate(scenarios)
+    def validate_fields(cls, scenarios, field_names):
+        super().validate_fields(scenarios, field_names)
         cls.check_domain(scenarios)
 
     @classmethod
-    def check_domain(cls, scenarios):
+    def check_domain(cls, scenarios, field_names=None):
         pool = Pool()
+        if field_names and not (field_names & {'model', 'domain'}):
+            return
         for scenario in scenarios:
             Model = pool.get(scenario.model)
             try:
@@ -205,10 +221,14 @@ class Scenario(Workflow, ModelSQL, ModelView):
 
         for scenario in scenarios:
             Model = pool.get(scenario.model)
-            sql_int = scenario.__class__.id.sql_type().base
             record = Record.__table__()
             cursor = Transaction().connection.cursor()
             domain = PYSONDecoder({}).decode(scenario.domain)
+            domain = [
+                domain,
+                ('marketing_party.marketing_scenario_unsubscribed',
+                    'not where', [('id', '=', scenario.id)]),
+                ]
             try:
                 query = Model.search(domain, query=True, order=[])
             except Exception:
@@ -218,10 +238,7 @@ class Scenario(Workflow, ModelSQL, ModelView):
                 continue
             cursor.execute(*(
                     query - record.select(
-                        Substring(
-                            record.record,
-                            Position(',', record.record) + Literal(1)
-                            ).cast(sql_int),
+                        Record.record.sql_id(record.record, Model),
                         where=record.scenario == scenario.id)))
 
             records = []
@@ -247,7 +264,7 @@ class Activity(ModelSQL, ModelView):
     "Marketing Activity"
     __name__ = 'marketing.automation.activity'
 
-    name = fields.Char("Name", required=True)
+    name = fields.Char("Name", translate=True, required=True)
     parent = fields.Reference(
         "Parent", [
             ('marketing.automation.scenario', "Scenario"),
@@ -269,7 +286,6 @@ class Activity(ModelSQL, ModelView):
         states={
             'invisible': ~Eval('event'),
             },
-        depends=['event'],
         help="Check to execute the activity "
         "if the event has not happened by the end of the delay.")
     on = fields.Function(fields.Selection([
@@ -287,10 +303,13 @@ class Activity(ModelSQL, ModelView):
 
     delay = fields.TimeDelta(
         "Delay",
+        domain=['OR',
+            ('delay', '=', None),
+            ('delay', '>=', TimeDelta()),
+            ],
         states={
             'required': Eval('negative', False),
             },
-        depends=['negative'],
         help="After how much time the action should be executed.")
 
     action = fields.Selection([
@@ -303,7 +322,6 @@ class Activity(ModelSQL, ModelView):
         states={
             'invisible': Eval('action') != 'send_email',
             },
-        depends=['action'],
         help="Leave empty to use the value defined in the configuration file.")
     email_title = fields.Char(
         "E-Mail Title",
@@ -312,7 +330,6 @@ class Activity(ModelSQL, ModelView):
             'invisible': Eval('action') != 'send_email',
             'required': Eval('action') == 'send_email',
             },
-        depends=['action'],
         help="The subject of the email.\n"
         "The Genshi syntax can be used "
         "with 'record' in the evaluation context.")
@@ -323,7 +340,6 @@ class Activity(ModelSQL, ModelView):
             'invisible': Eval('action') != 'send_email',
             'required': Eval('action') == 'send_email',
             },
-        depends=['action'],
         help="The HTML content of the E-mail.\n"
         "The Genshi syntax can be used "
         "with 'record' in the evaluation context.")
@@ -335,15 +351,14 @@ class Activity(ModelSQL, ModelView):
             "E-Mails Opened",
             states={
                 'invisible': Eval('action') != 'send_email',
-                },
-            depends=['action']), 'get_record_count')
+                }), 'get_record_count')
     email_clicked = fields.Function(
         fields.Integer(
             "E-Mails Clicked",
             states={
                 'invisible': Eval('action') != 'send_email',
-                },
-            depends=['action']), 'get_record_count')
+                }),
+        'get_record_count')
 
     @classmethod
     def __setup__(cls):
@@ -358,7 +373,7 @@ class Activity(ModelSQL, ModelView):
                     [(name, 'in', events + [None])],
                     domain)
             field.domain = [domain]
-            field.depends = ['parent_action']
+            field.depends.add('parent_action')
 
     @classmethod
     def view_attributes(cls):
@@ -441,14 +456,16 @@ class Activity(ModelSQL, ModelView):
         return count
 
     @classmethod
-    def validate(cls, activities):
-        super().validate(activities)
+    def validate_fields(cls, activities, fields_names):
+        super().validate_fields(activities, fields_names)
         for activity in activities:
-            activity.check_condition()
-            activity.check_email_title()
-            activity.check_email_template()
+            activity.check_condition(fields_names)
+            activity.check_email_title(fields_names)
+            activity.check_email_template(fields_names)
 
-    def check_condition(self):
+    def check_condition(self, fields_names=None):
+        if fields_names and 'condition' not in fields_names:
+            return
         if not self.condition:
             return
         try:
@@ -460,7 +477,9 @@ class Activity(ModelSQL, ModelView):
                     activity=self.rec_name,
                     exception=exception)) from exception
 
-    def check_email_template(self):
+    def check_email_template(self, fields_names=None):
+        if fields_names and 'email_template' not in fields_names:
+            return
         if not self.email_template:
             return
         try:
@@ -472,7 +491,9 @@ class Activity(ModelSQL, ModelView):
                     activity=self.rec_name,
                     exception=exception)) from exception
 
-    def check_email_title(self):
+    def check_email_title(self, fields_names=None):
+        if fields_names and 'email_title' not in fields_names:
+            return
         if not self.email_title:
             return
         try:
@@ -504,33 +525,18 @@ class Activity(ModelSQL, ModelView):
                 for child in self.children])
 
     def _email_recipient(self, record):
-        pool = Pool()
-        try:
-            Party = pool.get('party.party')
-        except KeyError:
-            Party = None
-        try:
-            Sale = pool.get('sale.sale')
-        except KeyError:
-            Sale = None
-
-        def get_party_email(party):
-            contact = party.contact_mechanism_get('email')
-            if contact and contact.email:
-                return _formataddr(
-                    contact.name or party.rec_name,
-                    contact.email)
-            return None
-
-        if Party and isinstance(record, Party):
-            return get_party_email(record)
-        elif Sale and isinstance(record, Sale):
-            return get_party_email(record.party)
+        party = record.marketing_party
+        contact = party.contact_mechanism_get('email')
+        if contact and contact.email:
+            return _formataddr(
+                contact.name or party.rec_name,
+                contact.email)
 
     def execute_send_email(
             self, record_activity, smtpd_datamanager=None, **kwargs):
         pool = Pool()
         WebShortener = pool.get('web.shortened_url')
+        Email = pool.get('ir.email')
         record = record_activity.record
 
         with Transaction().set_context(language=record.language):
@@ -599,7 +605,7 @@ class Activity(ModelSQL, ModelView):
         from_ = (config.get('marketing', 'email_from')
             or config.get('email', 'from'))
         msg = MIMEMultipart('alternative')
-        msg['From'] = translated.email_from or from_
+        set_from_header(msg, from_, translated.email_from or from_)
         msg['To'] = to
         msg['Subject'] = Header(title, 'utf-8')
         if html2text:
@@ -615,6 +621,16 @@ class Activity(ModelSQL, ModelView):
             sendmail_transactional(
                 from_, to_addrs, msg, datamanager=smtpd_datamanager)
 
+            email = Email(
+                recipients=to,
+                addresses=[{'address': a} for a in to_addrs],
+                subject=title,
+                resource=record.record,
+                marketing_automation_activity=self,
+                marketing_automation_record=record,
+                )
+            email.save()
+
     def email_context(self, record):
         return {
             'record': record.record,
@@ -636,8 +652,12 @@ class Record(ModelSQL, ModelView):
         required=True, ondelete='CASCADE')
     record = fields.Reference(
         "Record", selection='get_models', required=True)
-    blocked = fields.Boolean("Blocked")
-    uuid = fields.Char("UUID", readonly=True)
+    blocked = fields.Boolean(
+        "Blocked",
+        states={
+            'readonly': ~Eval('blocked', False),
+            })
+    uuid = fields.Char("UUID", readonly=True, strip=False)
 
     @classmethod
     def __setup__(cls):
@@ -650,6 +670,11 @@ class Record(ModelSQL, ModelView):
             ('uuid_unique', Unique(t, t.uuid),
                 'marketing_automation.msg_record_uuid_unique'),
             ]
+        cls._buttons.update({
+                'block': {
+                    'invisible': Eval('blocked', False),
+                    },
+                })
 
     @classmethod
     def default_uuid(cls):
@@ -681,27 +706,28 @@ class Record(ModelSQL, ModelView):
 
     @property
     def language(self):
-        pool = Pool()
-        try:
-            Party = pool.get('party.party')
-        except KeyError:
-            Party = None
-        try:
-            Sale = pool.get('sale.sale')
-        except KeyError:
-            Sale = None
-
-        if Party and isinstance(self.record, Party):
-            if self.record.lang:
-                return self.record.lang.code
-        elif Sale and isinstance(self.record, Sale):
-            if self.record.party.lang:
-                return self.record.party.lang.code
+        lang = self.record.marketing_party.lang
+        if lang:
+            return lang.code
 
     @dualmethod
+    @ModelView.button
     def block(cls, records):
+        pool = Pool()
+        Party = pool.get('party.party')
+
         cls.write(records, {'blocked': True})
 
+        parties = defaultdict(set)
+        for record in records:
+            if record.scenario.unsubscribable:
+                parties[record.record.marketing_party].add(record.scenario.id)
+        if parties:
+            Party.write(*sum((
+                        ([p], {'marketing_scenario_unsubscribed': [
+                                    ('add', s)]})
+                        for p, s in parties.items()), ()))
+
     def get_rec_name(self, name):
         if self.record:
             return self.record.rec_name
@@ -735,25 +761,22 @@ class RecordActivity(Workflow, ModelSQL, ModelView):
         "At",
         states={
             'readonly': Eval('state') != 'waiting',
-            },
-        depends=['state'])
+            })
     email_opened = fields.Boolean(
         "E-Mail Opened",
         states={
             'invisible': Eval('activity_action') != 'send_email',
-            },
-        depends=['activity_action'])
+            })
     email_clicked = fields.Boolean(
         "E-Mail Clicked",
         states={
             'invisible': Eval('activity_action') != 'send_email',
-            },
-        depends=['activity_action'])
+            })
     state = fields.Selection([
             ('waiting', "Waiting"),
             ('done', "Done"),
             ('cancelled', "Cancelled"),
-            ], "State", required=True, readonly=True)
+            ], "State", required=True, readonly=True, sort=False)
 
     @classmethod
     def __setup__(cls):
@@ -763,6 +786,11 @@ class RecordActivity(Workflow, ModelSQL, ModelView):
             ('activity_record_unique', Unique(t, t.activity, t.record),
                 'marketing_automation.msg_activity_record_unique'),
             ]
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state.in_(['waiting'])))
         cls._transitions |= set((
                 ('waiting', 'done'),
                 ('waiting', 'cancelled'),
